/*************************************************
misc.c
P. Clote, Jul 2004
Miscellaneous routines for reading in RNA either as
string from command line or FASTA file, etc.

NOTES:
1) I chose a[] to be an array of short int, rather than
   char, since a[0] is the length of the original RNA
   sequence, stored in a[1],...,a[ a[0] ]
   Thus this requires a special irintRNA() function.

*************************************************/




#ifndef __SPECTRUM_H__
#include "spectrum.h"
#endif
                                                                                
#define GUbond -1
#define AUbond -2
#define GCbond -3

void printRNA(short *a,short flag){
  //if flag=0 then don't print vertical bars
  int i, len;
  len = a[0];
  //print RNA
  for (i=1;i<=len;i++) printf("%c",a[i]);
  printf("\n");
  //now print separation markers every 10 spaces
  if (flag!=0 && len >= 10){
    for (i=1;i<=len;i++) {
      if (i%10==0) 
        printf("|");
      else
        printf(" ");
      }
    printf("\n");
    }
  }

char * constructString(short *a){
  int i, n;
  char *s; 

  n = a[0];
  s = (char *) calloc(n+1,sizeof(char));

  //print RNA
  for (i=0;i<n;i++) s[i] = (char) a[i+1];
  s[n] = '\0';
  return s;
  }




short *getRNA(char *flag, char *rna){
  //Function to retrieve RNA sequence, either input in command line
  //or as FASTA file
  FILE * infile;
  char ch;
  int i,n,len;
  static short a[NN0];  
    //store RNA sequence, static required to retain local variable contents
    //after function call

  /***** retrieve RNA sequence  **/
  if (strcmp("-s",flag) == 0)  {
    len = strlen(rna);
    a[0] = len;
    if (len>=NN0-2) {
      fprintf(stderr,"Length of RNA exceeds array size %d\n",NN0);
      exit(1);
      }
    for (i=0;i<len;i++) {
      a[i+1] = toupper(rna[i]);
      if (a[i+1] == 'T') a[i+1] = 'U';
      }
    //printRNA(a,1);
    }
  else { //read FASTA format file of RNA
    infile = fopen(rna,"r");
    ch=fgetc(infile);
    if (ch == '>') { //now remove FASTA comment
      while (ch != '\n') ch = fgetc(infile);
      }
      // This only works on Unix, since EOLN marker for DOS is \r\n
      // and of Macintosh is \r
      // At this point, ch is \n, since we have exited while loop
    a[0] = 0; //we'll need to return to set a[0] to be sequence length
    n=1; //current position of insert into RNA string
    while (ch != EOF && n < NN0){
      if (ch == '\n') {
        ch = fgetc(infile);
        if (ch == EOF) { a[0]=n-1; break; }
        }
      a[n] = toupper(ch);  n++;
      ch = fgetc(infile); 
      }
    if (n>=NN0){
      fprintf(stderr,"Length of RNA exceeds array size %d\n",NN0);
      exit(1);
      }
    a[0] = n-1;  //string length is n-1
    a[n+1] = '\0'; //string terminator
    //printRNA(a,1);
    fclose(infile);
    } 
  return a;
} 






void error(char *s){
   fprintf(stderr,"%s\n",s);
   exit(1);
}

int guPair(char x, char y){
   return ( x=='G'&&y=='U' || x=='U'&&y=='G' );
}

int auPair(char x, char y){
   return ( x=='A'&&y=='U' || x=='U'&&y=='A' );
}

int cgPair(char x, char y){
   return ( x=='C'&&y=='G' || x=='G'&&y=='C' );
}


int bondEnergy(int i, int j, char *s) {
  //returns the energy value of the pairing between
  //the ith and jth base in s
  if (i < strlen (s) && j < strlen (s)) {
      if (auPair(s[i],s[j]))
         return AUbond;
      else if (cgPair(s[i],s[j]))
         return CGbond;
      else if (guPair(s[i],s[j]))
         return GUbond;
      else
         return INF;
      }
    else
         error("Indices out of bounds for RNA sequence\n");
}




int basePair(char x, char y){
   return( auPair(x,y) || cgPair(x,y) || guPair(x,y) );
}


int watsonCrick(char x, char y){
   return( auPair(x,y) || cgPair(x,y) );
}


void printMatrix(MatrixType M, int NN){
  int n=NN;
  int i,j,k,s,b;
  for (k=0;k<MIN(KK-1,n/2);k++) {
    printf("k: %d ",k);
    printf("-------------------------------------\n");
    for (i=1;i<=n;i++) {
      for (j=i;j<=n;j++){
//        if (k>(j-i)/2) continue;
        printf("(%d,", BP(i,j,k));
        printf("%d,", VIS(i,j,k));
        printf("%d,",BS(i,j,k));
        printf("%d)\t",CASE(i,j,k,VIS(i,j,k),BS(i,j,k)));
        //WARNING: The case actually taken may use s!=VIS(i,j,k) and b!=BS(i,j,i)
        }
      printf("\n");
      }
  }
  printf("\n\n");
  for (k=0;k<MIN(KK-1,n/2);k++) {
    printf("k: %d ",k);
    printf("-------------------------------------\n");
    for (i=1;i<=n;i++) 
      for (j=i+1;j<=n;j++) 
        for (s=0;s<SS;s++) 
          for (b=0;b<BB;b++)
            if (BPs(i,j,k,s,b) != -1){
              printf("i:%d j:%d s:%d b:%d bp:%d partner:%d ",\
                    i,j,s,b,BPs(i,j,k,s,b),BPpartner(i,j,k,s,b));
              printf("lLt:%d lRt:%d ",\
                    LEVELleft(i,j,k,s,b),LEVELright(i,j,k,s,b));
              printf("visL:%d visR:%d ",\
                    VISleft(i,j,k,s,b),VISright(i,j,k,s,b));
              printf("bsL:%d bsR:%d case:%d\n",\
                    BSleft(i,j,k,s,b),BSright(i,j,k,s,b),CASE(i,j,k,s,b));
              }

  }
}


void printMatrixDouble(MatrixTypeDouble MM, int NN){
  int n=NN;
  int i,j,k,s,b;
  float n0;
  for (k=0;k<MIN(KK-1,n/2);k++) {
    printf("k: %d ",k);
    printf("-------------------------------------\n");
//    for (i=1;i<=n;i++) {
    for (i=1;i<=n-1;i++) {
      for (j=i;j<=n;j++){
        n0=0;
        for (s=0;s<SS;s++){
          for (b=0;b<BB;b++){
            n0 += NUMSUBOPT(i,j,k,s,b); 
            if ( NUMSUBOPT(i,j,k,s,b) > 0)
              printf("i:%d j:%d s:%d b:%d %lf\n",i,j,s,b,NUMSUBOPT(i,j,k,s,b));
            }
          }
        printf("\nNumSubOpt(%d,%d,%d)=%lf\n",i,j,k,n0);
        }
      }
  }
}


short nuclAppears(short *a,int i, int j){
  //Function returns integer s which encodes which nucleotides
  //A,C,G,U appear in a[i],...,a[j].
  //BIT(s,0) is 1/0 depending on whether 'A' in a[i],...,a[j]
  //BIT(s,1) is 1/0 depending on whether 'C' in a[i],...,a[j]
  //BIT(s,2) is 1/0 depending on whether 'G' in a[i],...,a[j]
  //BIT(s,3) is 1/0 depending on whether 'U' in a[i],...,a[j]
  int k, s = 0, found=0;
  for (k=i;k<=j;k++)
    if ('A' == a[k]){
      s += 1;
      break;
    }
  for (k=i;k<=j;k++)
    if ('C' == a[k]){
      s += 2;
      break;
    }
  for (k=i;k<=j;k++)
    if ('G' == a[k]){
      s += 4;
      break;
    }
  for (k=i;k<=j;k++)
    if ('U' == a[k]){
      s += 8;
      break;
    }
  return s;
}


int addElement(int x,int s){
  //add element x to set s
  char text[256];

  sprintf(text,"%c in set coded by %d\n",x,s);
  if ( in(x,s) ) error(text);
  if (x=='A') return s+1;
  if (x=='C') return s+2;
  if (x=='G') return s+4;
  if (x=='U') return s+8;
}



int initializeVis(short *a, int i,int j){
  //code correct even if i <= j <= i+THRESHOLD
  int m;
  if (j-i<=THRESHOLD) return 0;
  m = j-THRESHOLD-1;
  return nuclAppears(a,i,m);

}

int initializeBitString(int i,int j){
  //code correct even if i <= j <= i+THRESHOLD
  int k,m,b=0;
  m=MAX(0,MIN(j-i+1,THRESHOLD+1));
  return m;
}

int joinSandB(short *a, int s, int b, int i, int j){
  int r,s0;
  s0 = s;
  for (r=0;r<b;r++)
    if ( !in(a[j-r],s0) ) s0=addElement(a[j-r],s0);
  return s0;
}


int canBasePairWithElementInSet(int x, int s){
  //Compute whether x can basepair with an element in set s
  if ( basePair(x,'A') && BIT(s,0)==1 )  return 1;
  if ( basePair(x,'C') && BIT(s,1)==1 )  return 1;
  if ( basePair(x,'G') && BIT(s,2)==1 )  return 1;
  if ( basePair(x,'U') && BIT(s,3)==1 )  return 1;
  return 0;
}


int CASEfunction(MatrixType M, int i,int j, int k, int s, int b){
  int partner = BPpartner(i,j,k,s,b);
  if (partner==0) return 3;
  else if (partner==i) return 1;
  else if (i<partner && partner<i) return 2;
  else return -1;
}

